import numpy as np

from .utils import assert_in, first, one
from .arrays import ParameterArray as parray
from .arrays import UncertainParameterArray as uparray


def make_deltas_parray(*, stdzr, scale, **deltas):
    """Make a parray containing standardized differences for each dimension.

    Primarily used to create bounds for the lengthscale hyperparameters of a GP
    model.
    """
    assert_in("scale", scale, ["transformed", "standardized", "natural"])
    match scale:
        case "transformed":
            deltas = {
                dim: [stdzr.untransform(dim, [v, v * 2]) if v is not None else None for v in vs]
                for dim, vs in deltas.items()
            }
        case "standardized":
            deltas = {
                dim: [stdzr.unstdz(dim, [v, v * 2]) if v is not None else None for v in vs]
                for dim, vs in deltas.items()
            }
        case "natural":
            deltas = {dim: [[v, v * 2] if v is not None else None for v in vs] for dim, vs in deltas.items()}

    deltas = {
        dim: [np.diff(stdzr.stdz(dim, v)) if v is not None else [np.nan] for v in vs] for dim, vs in deltas.items()
    }
    ls_bounds = parray(**deltas, stdzr=stdzr, stdzd=True)
    return ls_bounds


def stack(array_list, axis=0, **kwargs):
    ndims = {pa.ndim for pa in array_list}
    if ndims == {1}:
        return hstack(array_list)
    types = {type(pa) for pa in array_list}
    if not len(types) == 1:
        raise ValueError("Arrays are not all of the same type.")
    cls = one(types)
    if cls == parray:
        return _stack_parray(array_list, **kwargs)
    elif cls == uparray:
        all_names = [upa.name for upa in array_list]
        if not len(set(all_names)) == 1:
            raise ValueError("Arrays do not have the same name.")
    else:
        raise ValueError(f"Unknown array type: {cls}")
    new = np.stack(array_list, axis=axis, **kwargs)
    stdzr = _get_stdzr(array_list)
    return cls(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


def vstack(array_list, **kwargs):
    types = {type(pa) for pa in array_list}
    if not len(types) == 1:
        raise ValueError("Arrays are not all of the same type.")
    cls = one(types)
    if cls == parray:
        return _vstack_parray(array_list, **kwargs)
    elif cls == uparray:
        all_names = [upa.name for upa in array_list]
        if not len(set(all_names)) == 1:
            raise ValueError("Arrays do not have the same name.")
    else:
        raise ValueError(f"Unknown array type: {cls}")
    new = np.vstack(array_list, **kwargs)
    stdzr = _get_stdzr(array_list)
    return cls(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


def hstack(array_list, **kwargs):
    types = {type(pa) for pa in array_list}
    if not len(types) == 1:
        raise ValueError("Arrays are not all of the same type.")
    cls = one(types)
    if cls == parray:
        return _hstack_parray(array_list, **kwargs)
    elif cls == uparray:
        all_names = [upa.name for upa in array_list]
        if not len(set(all_names)) == 1:
            raise ValueError("Arrays do not have the same name.")
    else:
        raise ValueError(f"Unknown array type: {cls}")
    new = np.hstack(array_list, **kwargs)
    stdzr = _get_stdzr(array_list)
    return cls(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


def _stack_parray(array_list, axis=0, **kwargs):
    _check_parrays_compatible(array_list)
    new = np.stack(array_list, axis=axis, **kwargs)
    stdzr = _get_stdzr(array_list)
    return parray(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


def _vstack_parray(array_list, **kwargs):
    _check_parrays_compatible(array_list)
    new = np.vstack(array_list, **kwargs)
    stdzr = _get_stdzr(array_list)
    return parray(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


def _hstack_parray(array_list, **kwargs):
    _check_parrays_compatible(array_list)
    new = np.hstack(array_list, **kwargs)
    stdzr = _get_stdzr(array_list)
    return parray(**{dim: new[dim] for dim in new.dtype.names}, stdzr=stdzr)


def _check_parrays_compatible(array_list):
    all_names = [tuple(pa.names) for pa in array_list]
    if not len(set(all_names)) == 1:
        raise ValueError("Arrays do not have the same names.")
    if not all(names == first(all_names) for names in all_names):
        raise ValueError("Arrays names are not in the same order.")


def _get_stdzr(array_list):
    stdzr = first(array_list).stdzr
    if not all(a.stdzr is stdzr for a in array_list):
        raise ValueError("Arrays do not have the same standardizer.")
    return stdzr



"""Tests behavior of structured arrays"""

import matplotlib.pyplot as plt
import numpy as np

from gumbi import *


def test_1d_array_plotting():
    stdzr = Standardizer(
        c={"μ": -5.30, "σ": 0.582},
        d={"μ": -0.307, "σ": 0.158},
        z={"μ": 2, "σ": 2},
        log_vars=["d", "c"],
        logit_vars=["z"],
    )
    m_pa = parray(m=np.arange(1, 5) / 10, stdzr=stdzr)
    r_upa = uparray("d", np.arange(1, 5) / 10 + 0.5, np.arange(1, 5) / 100 * 2, stdzr=stdzr)
    ParrayPlotter(m_pa, r_upa.t).plot()
    ParrayPlotter(m_pa, r_upa.z).plot()
    ParrayPlotter(m_pa.t, r_upa.t).plot()
    ParrayPlotter(m_pa.z, r_upa.z).plot()
    ParrayPlotter(m_pa, r_upa, y_scale="standardized").plot()
    ParrayPlotter(m_pa, r_upa, y_scale="transformed").plot()
    ParrayPlotter(m_pa, r_upa, y_scale="transformed", x_scale="standardized").plot()
    ParrayPlotter(m_pa, r_upa, y_scale="standardized", x_scale="transformed").plot()
    ParrayPlotter(m_pa, r_upa, y_scale="transformed", x_scale="standardized").plot()
    ParrayPlotter(m_pa, r_upa.z, y_tick_scale="standardized").plot()
    ParrayPlotter(m_pa, r_upa.t, y_tick_scale="transformed").plot()
    ParrayPlotter(m_pa.z, r_upa, x_tick_scale="standardized").plot()
    ParrayPlotter(m_pa.t, r_upa, x_tick_scale="transformed").plot()


def test_2d_array_plotting():
    stdzr = Standardizer(c={"μ": -5.30, "σ": 0.582}, d={"μ": -0.307, "σ": 0.158}, log_vars=["d", "c"])
    c = np.arange(1, 10, 0.25)
    d = np.arange(1, 10, 0.25)
    c, d = np.meshgrid(c, d)
    Z = np.sin(np.sqrt((c - 5) ** 2 + (d - 5) ** 2))
    mrz = ParameterArray(c=c, d=d, z=Z, stdzr=stdzr)
    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"])
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz["c"].t, mrz["d"].t, mrz["z"].t)
    cs = pp(plt.contourf)
    pp = ParrayPlotter(mrz["c"].z, mrz["d"].z, mrz["z"].z)
    cs = pp(plt.contourf)
    pp = ParrayPlotter(
        mrz["c"],
        mrz["d"],
        mrz["z"],
        x_scale="standardized",
        y_scale="standardized",
        z_scale="standardized",
    )
    cs = pp(plt.contourf)
    pp = ParrayPlotter(
        mrz["c"],
        mrz["d"],
        mrz["z"],
        x_scale="transformed",
        y_scale="transformed",
        z_scale="transformed",
    )
    cs = pp(plt.contourf)
    pp = ParrayPlotter(
        mrz["c"].z,
        mrz["d"].t,
        mrz["z"],
        x_tick_scale="standardized",
        y_tick_scale="transformed",
    )
    cs = pp(plt.contourf)
    pp = ParrayPlotter(
        mrz["c"].t,
        mrz["d"].z,
        mrz["z"],
        x_tick_scale="transformed",
        y_tick_scale="standardized",
    )
    cs = pp(plt.contourf)

    cs


def test_colorbar():
    c = np.arange(1, 10, 0.25)
    d = np.arange(1, 10, 0.25)
    c, d = np.meshgrid(c, d)
    Z = np.sin(np.sqrt((c - 5) ** 2 + (d - 5) ** 2))
    stdzr = Standardizer(
        c={"μ": -5, "σ": 0.5},
        d={"μ": -0.3, "σ": 0.15},
        z={"μ": 2, "σ": 2},
        log_vars=["d", "c"],
        logit_vars=["z"],
    )
    mrz = ParameterArray(c=c, d=d, z=Z**2, stdzr=stdzr)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"])
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"].t)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"].z)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"].t)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"], z_scale="transformed")
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"].z)
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    pp = ParrayPlotter(mrz["c"], mrz["d"], mrz["z"], z_scale="standardized")
    cs = pp(plt.contourf)
    cbar = pp.colorbar(cs)

    cbar



import pathlib as pl

import numpy as np
import pandas as pd
import pytest

from gumbi import DataSet, Standardizer, WideData

test_dir = pl.Path(__file__).resolve().parent
test_data = test_dir / "test_data"

example_stdzr = {
    "a": {"μ": -0.762, "σ2": 1.258**2},
    "b": {"μ": -0.0368, "σ2": 0.351**2},
    "c": {"μ": -5.30, "σ2": 0.582**2},
    "d": {"μ": -0.307, "σ2": 0.158**2},
    "e": {"μ": -1.056, "σ2": 0.398**2},
    "f": {"μ": 3.34, "σ2": 0.1501**2},
    "X": {"μ": -0.282, "σ2": 1**2},
    "Y": {"μ": 4.48, "σ2": 0.75**2},
    "lg10_Z": {"μ": 5, "σ2": 2**2},
}

log_vars = ["d", "f", "b", "c", "Y"]
logit_vars = ["e", "X"]

################################################################################
# Standardizer
################################################################################


def test_stdz():
    s = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)
    nat_defaults = {p: s.untransform(p, v["μ"]) for p, v in example_stdzr.items()}
    assert np.allclose(
        [s.stdz(p, nat_defaults[p]) for p, v in nat_defaults.items()], 0
    ), "Standardization failed"
    assert np.allclose(
        [s.unstdz(p, s.stdz(p, nat_defaults[p])) for p, v in example_stdzr.items()],
        [v for v in nat_defaults.values()],
    ), "Round-trip standardization failed"

    stdzr = Standardizer(x={"μ": 1, "σ2": 0.1}, d={"μ": 0, "σ2": 0.1}, log_vars=["d"])
    assert stdzr.transform("x", μ=1) == 1
    assert stdzr.stdz("x", 1) == 0.0
    assert stdzr.unstdz("x", 0) == 1.0
    assert np.isclose(stdzr.stdz("x", 1 + 0.1**0.5), 1.0)
    assert np.isclose(stdzr.unstdz("x", 1), 1 + 0.1**0.5)
    assert stdzr.stdz("d", 1) == 0.0
    assert np.isclose(stdzr.stdz("d", np.exp(0.1**0.5)), 1.0)

    assert stdzr.transform("x", μ=1, σ2=0.1) == (1, 0.1)
    assert stdzr.stdz("x", 1, 0.1) == (0.0, 1.0)
    assert stdzr.stdz("d", 1, 0.1) == (0.0, 1.0)
    assert stdzr.transform("d", 1, 0.1) == (0.0, 0.1)

    x_series = pd.Series(np.arange(1, 5), name="x")
    x_stdzd = np.array([0.0, 3.162278, 6.324555, 9.486833])
    assert np.allclose(stdzr.stdz(x_series).values, x_stdzd)

    r_series = pd.Series(np.arange(1, 5), name="d")
    r_stdzd = np.array([0.0, 2.19192384, 3.4741171, 4.38384769])
    assert np.allclose(stdzr.stdz(r_series).values, r_stdzd)


################################################################################
# DataSet
################################################################################


@pytest.fixture
def example_dataset():
    df = pd.read_pickle(test_data / "estimates_test_data.pkl")
    ds = DataSet.from_tidy(
        df, names_column="Parameter", log_vars=["Y", "c", "b"], logit_vars=["X", "e"]
    )
    assert ds.wide is not None
    assert ds.wide.z is not None
    assert ds.tidy is not None
    assert ds.tidy.z is not None
    return ds


def test_tidy_z(example_dataset):
    ds = example_dataset
    assert ds.tidy.z.shape == ds.tidy.shape
    assert np.allclose(
        [
            ds.tidy.z[ds.tidy.z.Parameter == p]["Value"].mean()
            for p in ds.tidy.z.Parameter.unique()
        ],
        0,
    )


def test_wide_io(example_dataset):
    ds = example_dataset

    wide_out = ds.wide
    wide_in_wd = WideData(
        wide_out, outputs=ds.outputs, log_vars=["Y", "c", "b"], logit_vars=["X", "e"]
    )
    wide_in_ds = DataSet(
        wide_out, outputs=ds.outputs, log_vars=["Y", "c", "b"], logit_vars=["X", "e"]
    )

    pd.testing.assert_frame_equal(wide_in_wd, wide_out)
    pd.testing.assert_frame_equal(wide_in_wd, wide_in_ds.wide)
    pd.testing.assert_frame_equal(ds.wide, wide_in_ds.wide)
    pd.testing.assert_frame_equal(ds.wide, wide_in_wd)

    ds.wide = wide_out.drop(0)
    pd.testing.assert_frame_equal(ds.wide, wide_out.drop(0))
    pd.testing.assert_frame_equal(ds.wide, wide_in_wd.drop(0))



from gumbi import Regressor
from gumbi.utils import listify, assert_in
import numpy as np

import torch
from torch.nn.functional import pdist
from botorch import fit_gpytorch_mll
from botorch.models import SingleTaskGP, MixedSingleTaskGP, HeteroskedasticSingleTaskGP, MultiTaskGP, KroneckerMultiTaskGP
from botorch.models.transforms.input import Normalize
from botorch.acquisition.objective import GenericMCObjective
from botorch.acquisition import qLogNoisyExpectedImprovement
from botorch.acquisition.multi_objective.logei import qLogNoisyExpectedHypervolumeImprovement
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.optim.optimize import optimize_acqf
from botorch.utils.transforms import unnormalize as bt_unnormalize
from botorch.utils.transforms import normalize as bt_normalize
from gpytorch.kernels import RBFKernel, MaternKernel
from gpytorch.mlls import ExactMarginalLogLikelihood
import gpytorch

from warnings import simplefilter
from linear_operator.utils.warnings import NumericalWarning

import warnings


class BotorchGP(Regressor):

    def __init__(self, *args, gpu=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() and gpu else "cpu"
        )
        self.model = None
        self.mll = None
        self.structure = None

    def fit(
        self,
        outputs=None,
        *,
        linear_dims=None,
        continuous_dims=None,
        continuous_levels=None,
        continuous_coords=None,
        categorical_dims=None,
        categorical_levels=None,
        additive=False,
        seed=None,
        continuous_kernel="ExpQuad",
        period=None,
        heteroskedastic_inputs=False,
        heteroskedastic_outputs=False,
        sparse=False,
        n_u=100,
        ARD=False,
        ls_bounds=None,
        spec_kwargs=None,
        build_kwargs=None,
        multitask_kernel=None,
        **kwargs,
    ):

        if additive:
            raise NotImplementedError(
                "Additive models are not yet supported in BotorchGP."
            )
        if period is not None:
            raise NotImplementedError(
                "Periodic kernels are not yet supported in BotorchGP."
            )
        if linear_dims is not None:
            warnings.warn(
                "Linear dimensions are not yet supported in BotorchGP and will be ignored."
            )
        if categorical_dims is not None and len(listify(outputs)) > 1:
            raise NotImplementedError(
                "Categorical dimensions with multiple outputs are not yet supported in BotorchGP."
            )

        self.specify_model(
            outputs=outputs,
            linear_dims=linear_dims,
            continuous_dims=continuous_dims,
            continuous_levels=continuous_levels,
            continuous_coords=continuous_coords,
            categorical_dims=categorical_dims,
            categorical_levels=categorical_levels,
            additive=additive,
            **(spec_kwargs or {}),
        )

        self.build_model(
            seed=seed,
            continuous_kernel=continuous_kernel,
            period=period,
            heteroskedastic_inputs=heteroskedastic_inputs,
            heteroskedastic_outputs=heteroskedastic_outputs,
            sparse=sparse,
            n_u=n_u,
            ARD=ARD,
            ls_bounds=ls_bounds,
            multitask_kernel=multitask_kernel,
            **(build_kwargs or {}),
        )

        return self.fit_model(**kwargs)

    def build_model(
        self,
        *,
        continuous_kernel="ExpQuad",
        period=None,
        heteroskedastic_inputs=False,
        heteroskedastic_outputs=True,
        sparse=False,
        n_u=100,
        ARD=False,
        ls_bounds=None,
        seed=None,
        mass=0.98,
        multitask_kernel=None
    ):

        if period is not None:
            raise NotImplementedError(
                "Periodic kernels are not yet supported in BotorchGP."
            )
        if sparse:
            raise NotImplementedError(
                "Sparse models are not yet supported in BotorchGP."
            )
        if heteroskedastic_inputs:
            raise NotImplementedError(
                "Heteroskedasticity over inputs is not yet implemented in BotorchGP."
            )
        if heteroskedastic_outputs:
            raise NotImplementedError(
                "Heteroskedasticity over inputs is not yet implemented in BotorchGP."
            )
        if ls_bounds is not None:
            raise NotImplementedError(
                "User-defined lengthscale bounds are not yet implemented in BotorchGP."
            )
        if mass != self.build_model.__kwdefaults__["mass"]:
            warnings.warn(f"`mass` keyword ignored in {self.__class__.__name__}")
            
        if multitask_kernel is not None:
            multitask_kernel = multitask_kernel.capitalize()
        assert_in('multitask_kernel', multitask_kernel, [None, 'Kronecker','Hadamard'])

        X, y = self.get_shaped_data("mean")

        D_in = len([d for d in self.dims if d != self.out_col])
        D_out = len(self.outputs)
        if D_out == 1:
            assert X.shape[1] == D_in
        else:
            assert X.shape[1] == D_in + 1

        seed = self.seed if seed is None else seed
        self.seed = seed
        self.continuous_kernel = continuous_kernel
        self.heteroskedastic_inputs = heteroskedastic_inputs
        self.heteroskedastic_outputs = heteroskedastic_outputs
        self.sparse = sparse
        self.n_u = n_u
        self.latent = False

        Yvar = None
            
        if D_out == 1:
            cat_dims = [d for d in self.categorical_dims if d != self.out_col]
            n_cats = len(cat_dims)
            if n_cats == 0:
                if heteroskedastic_inputs:
                    self.structure = 'HeteroskedasticSingleTask'
                    self.model = HeteroskedasticSingleTaskGP(
                        X,
                        y,
                        train_Yvar=Yvar,
                        # covar_module=self._get_kernel(
                        #     continuous_kernel, ARD, ls_bounds=ls_bounds
                        # ),
                        input_transform=Normalize(d=D_in),
                    ).to(self.device)
                else:
                    self.structure = 'SingleTask'
                    self.model = SingleTaskGP(
                        X,
                        y,
                        train_Yvar=Yvar,
                        covar_module=self._get_kernel(
                            continuous_kernel, ARD, ls_bounds=ls_bounds
                        ),
                        input_transform=Normalize(d=D_in),
                    ).to(self.device)
            else:
                self.structure = 'MixedSingleTaskGP'
                cat_idxs = [self.dims.index(dim) for dim in cat_dims]
                self.model = MixedSingleTaskGP(
                    X,
                    y,
                    cat_dims=cat_idxs,
                    train_Yvar=Yvar,
                    cont_kernel_factory=self._get_kernel_factory(
                        continuous_kernel, ARD, ls_bounds=ls_bounds
                    ),
                    input_transform=Normalize(d=D_in),
                ).to(self.device)
        else:
            # Decide between Kronecker and Hadamard structure
            Xs, ys = self.get_separated_data("mean")
            kronecker_possible = all([torch.allclose(Xs[0], x) for x in Xs[1:]])
            kronecker_default = D_out > 2
            force_kronecker = multitask_kernel == 'Kronecker'
            if kronecker_possible and (kronecker_default or force_kronecker):
                # Kronecker structure: all outputs observed at every input point
                # Allows performance improvements for high-dimensional outputs
                self.structure = 'KroneckerMultiTaskGP'
                self.model = KroneckerMultiTaskGP(
                    Xs[0],
                    torch.cat(ys, dim=1),
                    data_covar_module=self._get_kernel(
                        continuous_kernel, ARD, ls_bounds=ls_bounds
                    ),
                    input_transform=Normalize(d=D_in),
                ).to(self.device)
            else:
                # Hadamard structure: outputs observed at different input points
                self.structure = 'MultiTaskGP'
                self.model = MultiTaskGP(
                    X,
                    y,
                    task_feature=-1,
                    train_Yvar=Yvar,
                    covar_module=self._get_kernel(
                        continuous_kernel, ARD, ls_bounds=ls_bounds
                    ),
                    input_transform=Normalize(d=D_in + 1),
                ).to(self.device)

        self.mll = ExactMarginalLogLikelihood(self.model.likelihood, self.model)

        return self.model, self.mll

    def get_shaped_data(self, metric="mean", dropna=True):
        X, y = super().get_shaped_data(metric=metric, dropna=dropna)

        X = torch.tensor(X).double().to(self.device)
        y = torch.tensor(y).double().to(self.device)
        return X, y.unsqueeze(-1)

    def get_separated_data(self, metric="mean", dropna=True):
        X, y = self.get_shaped_data(metric=metric, dropna=dropna)
        if len(self.outputs) == 1:
            return [X], [y]
        else:
            idxs = torch.unique(X[:, -1])
            Xs, ys = [], []
            for i in idxs:
                idx = X[:, -1] == i
                Xs.append(X[idx, :-1])
                ys.append(y[idx])
            return Xs, ys
        
    def _get_kernel_factory(self, kernel_type, ard, ls_bounds=None):

        # Apply lengthscale constraints if provided
        default_lower, default_upper = self._get_default_bounds(ard)
        if ls_bounds is not None:
            lower, upper = ls_bounds
            lower = max(lower, default_lower)
            upper = min(upper, default_upper)
        else:
            lower, upper = default_lower, default_upper

        lengthscale_constraint = gpytorch.constraints.Interval(lower, upper)

        
        def kernel_factory(*, batch_shape=None, ard_num_dims=None, active_dims=None):

            match kernel_type:
                case "RBF" | "ExpQuad":
                    base_kernel = RBFKernel(
                        lengthscale_constraint=lengthscale_constraint,
                        batch_shape=batch_shape,
                        ard_num_dims=ard_num_dims,
                        active_dims=active_dims,
                    )
                case "Matern32" | "Matern3/2":
                    base_kernel = MaternKernel(
                        nu=1.5,
                        lengthscale_constraint=lengthscale_constraint,
                        batch_shape=batch_shape,
                        ard_num_dims=ard_num_dims,
                        active_dims=active_dims,
                    )
                case "Matern52" | "Matern5/2":
                    base_kernel = MaternKernel(
                        nu=2.5,
                        lengthscale_constraint=lengthscale_constraint,
                        batch_shape=batch_shape,
                        ard_num_dims=ard_num_dims,
                        active_dims=active_dims,
                    )
                case _:
                    raise ValueError(
                        "Invalid kernel type. Choose 'RBF', 'ExpQuad', 'Matern32', 'Matern3/2', 'Matern5/2', or 'Matern52'."
                    )
            return base_kernel
        return kernel_factory

    def _get_kernel(self, continuous_kernel, ard, ls_bounds=None):
        """
        Returns the appropriate kernel based on user settings.
        """
                
        kernel_factory = self._get_kernel_factory(continuous_kernel, ard, ls_bounds=ls_bounds)
        D_in = len([d for d in self.dims if d != self.out_col])
        base_kernel = kernel_factory(ard_num_dims=D_in if ard else None)

        return base_kernel

    def _get_default_bounds(self, ard: bool):
        def _get_default_lower_upper(x):

            distances = pdist(x)
            distinct = distances != 0

            default_lower = distances[distinct].min() if sum(distinct) > 0 else 0.01
            default_upper = distances[distinct].max() if sum(distinct) > 0 else 1

            return default_lower, default_upper

        Xs, _ = self.get_separated_data("mean")
        X = torch.cat(Xs, dim=0)

        if ard:
            lengthscale_bounds = torch.stack(
                [torch.tensor(_get_default_lower_upper(X_.unsqueeze(1))) for X_ in X.T]
            ).T
            lower, upper = lengthscale_bounds.cpu().numpy().squeeze().tolist()
        else:
            lengthscale_bounds = torch.tensor(_get_default_lower_upper(X))
            lower, upper = lengthscale_bounds.cpu().numpy().squeeze().tolist()

        return lower, upper

    def fit_model(self, **kwargs):
        """
        Fits the GP model using exact marginal log likelihood.
        """
        torch.manual_seed(self.seed)
        return fit_gpytorch_mll(self.mll, **kwargs)

    def predict(self, points_array, *, with_noise=True, additive_level="total"):
        """
        Make predictions at supplied points using specified gp

        Parameters
        ----------
        points_array : np.ndarray
            Array of points at which to make predictions
        with_noise : bool, optional
            Whether to include noise in the predictions, by default True

        Returns
        -------
        mean, variance : tuple of np.ndarray
            Mean and variance of predictions at each supplied points
        """
        points_tensor = torch.tensor(points_array).float().to(self.device)

        if additive_level != "total":
            raise NotImplementedError(
                "Prediction for additive sublevels is not yet supported."
            )

        self.model.eval()
        self.model.likelihood.eval()

        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            t_points_tensor = self.model.input_transform.transform(points_tensor)
            predictive_distribution = self.model.likelihood(self.model(t_points_tensor))
            self.predictive_distribution = predictive_distribution
            if with_noise:
                # The likelihood will add noise to the posterior
                mean = predictive_distribution.mean
                variance = predictive_distribution.stddev**2
                # lower, upper = predictive_distribution.confidence_region()
            else:
                posterior = self.model.posterior(points_tensor)
                mean = posterior.mean
                variance = posterior.variance

        return mean.cpu().numpy(), variance.cpu().numpy()

    def predict_points(self, points, output=None, with_noise=True, **kwargs):
        """Make predictions at supplied points

        Parameters
        ----------
        points : ParameterArray
            1-D ParameterArray vector of coordinates for prediction, must have one layer per ``self.dims``
        output : str or list of str, optional
            Variable for which to make predictions
        with_noise : bool, default True
            Whether to incorporate aleatoric uncertainty into prediction error
        **kwargs
            Additional keyword arguments passed to subclass-specific :meth:`predict` method

        Returns
        -------
        prediction : UncertainParameterArray
            Predictions as a `uparray`
        """

        # TODO: add convenience method for predicting at a single point

        output = self._parse_prediction_output(output)
        D_out = len(output)
        points_array, tall_points, param_coords = self._prepare_points_for_prediction(
            points, output=output
        )
        
        if self.structure == 'KroneckerMultiTaskGP':
            # Kronecker structure: all outputs observed at every input point
            points_array = points_array[points_array[:,-1]==0, :-1]

        # Prediction means and variance as a list of numpy vectors
        pred_mean, pred_variance = self.predict(
            points_array, with_noise=with_noise, **kwargs
        )
        self.predictions_X = points
        
        if self.structure == 'KroneckerMultiTaskGP':
            pred_mean

        # Store predictions in appropriate structured array format
        if D_out == 1:
            # Predicting one output, return an UncertainParameterArray
            self.predictions = self.uparray(
                output[0], pred_mean, pred_variance, stdzd=True
            )
        else:
            # Predicting multiple parameters, return an MVUncertainParameterArray First split prediction into
            # UncertainParameterArrays
            uparrays = []
            out_coords = {v:k for k, v in self.categorical_coords[self.out_col].items()}
            for i in range(D_out):
                if self.structure != 'KroneckerMultiTaskGP':
                    # idx = (tall_points[self.out_col].values() == param_coords[i]).squeeze()
                    idx = points_array[:,-1] == i
                    μ = pred_mean[idx]
                    σ2 = pred_variance[idx]
                    name = out_coords[i]
                else:
                    μ = pred_mean[:,i]
                    σ2 = pred_variance[:,i]
                    name = out_coords[i]
                uparrays.append(self.uparray(name, μ, σ2, stdzd=True))

            # Calculate the correlation matrix from the hyperparameters of the coregion kernel
            with (
                torch.no_grad(),
                gpytorch.settings.fast_pred_var(),
                warnings.catch_warnings(),
            ):
                simplefilter("ignore", category=NumericalWarning)
                if self.structure == 'KroneckerMultiTaskGP':
                    task_covar_module = self.model.covar_module.task_covar_module
                else:
                    task_covar_module = self.model.task_covar_module

                W = task_covar_module.covar_factor
                κ = task_covar_module.var

                cov = W @ W.T + torch.diag(κ)

                σ_task = torch.sqrt(torch.diag(cov))

                σ_outer = σ_task.unsqueeze(0) * σ_task.unsqueeze(1)  # Outer product
                cor = cov / σ_outer

            cor = cor.cpu().numpy()

            # Store predictions as MVUncertainParameterArray
            self.predictions = self.mvuparray(*uparrays, cor=cor)

        return self.predictions

    def propose(
        self,
        *,
        q=1,
        bounds=None,
        maximize=True,
        num_restarts=10,
        raw_samples=100,
        mc_samples=256,
        seed=None,
        ref_point=None,
        sequential=False,
    ):
        """
        Optimizes the acquisition function (Expected Improvement) to propose new candidates.
        """

        seed = seed if seed is not None else self.seed
        torch.manual_seed(seed)

        qmc_sampler = SobolQMCNormalSampler(
            sample_shape=torch.Size([mc_samples]), seed=seed
        )

        X, y = self.get_shaped_data("mean")
        D_out = len(self.outputs)

        if bounds is None:
            if D_out == 1:
                bounds = torch.stack([X.min(0).values, X.max(0).values]).to(self.device)
            else:
                bounds = torch.stack(
                    [X[:, :-1].min(0).values, X[:, :-1].max(0).values]
                ).to(self.device)

        if D_out == 1:
            # Construct the objective function for the acquisition function

            def identity(samples, X=None):
                return samples.squeeze(-1)

            def neg_identity(samples, X=None):
                return -identity(samples, X)  # Negate the samples for minimization

            if maximize:
                objective = GenericMCObjective(identity)
            else:
                objective = GenericMCObjective(neg_identity)
            
            with warnings.catch_warnings():
                simplefilter("ignore", category=NumericalWarning)

                # Expected Improvement acquisition function
                acq_func = qLogNoisyExpectedImprovement(
                    model=self.model,
                    X_baseline=X,
                    sampler=qmc_sampler,
                    objective=objective,
                )
        else:

            Xs, ys = self.get_separated_data("mean")
            X_baseline = torch.unique(torch.cat(Xs, dim=0), dim=0)

            if ref_point is None:
                ref_point = []
                for y_ in ys:
                    if maximize:
                        ref_point.append(y_.min().item() - 1e-3)
                    else:
                        ref_point.append(y_.max().item() + 1e-3)

            self.ref_point = ref_point
            
            with warnings.catch_warnings():
                simplefilter("ignore", category=NumericalWarning)

                # Expected Hyper-Volume Improvement acquisition function
                acq_func = qLogNoisyExpectedHypervolumeImprovement(
                    model=self.model,
                    ref_point=torch.tensor(ref_point),  # use known reference point
                    X_baseline=bt_normalize(X_baseline, bounds),
                    prune_baseline=True,  # prune baseline points that have estimated zero probability of being Pareto optimal
                    sampler=qmc_sampler,
                    cache_root=False,
                )

                bounds = bt_normalize(bounds, bounds)

        with warnings.catch_warnings():
            simplefilter("ignore", category=NumericalWarning)

            # Optimize acquisition function to propose new candidates
            candidates, acqf_values = optimize_acqf(
                acq_function=acq_func,
                bounds=bounds,
                q=q,  # Batch size
                num_restarts=num_restarts,
                raw_samples=raw_samples,
                sequential=sequential,
            )

        if D_out > 1:
            # Unnormalize the candidates
            candidates = bt_unnormalize(candidates, bounds)

        candidates, acqf_values = candidates.cpu().numpy(), acqf_values.cpu().numpy()
        candidates = self.parray(
            **{dim: values for dim, values in zip(self.continuous_dims, candidates.T)},
            stdzd=True,
        )
        return candidates, acqf_values


import os
import pathlib as pl

style_dir = pl.Path(__file__).resolve().parent

futura = style_dir / "futura_presentation.mplstyle"
breve = style_dir / "breve_presentation.mplstyle"
default = style_dir / "presentation.mplstyle"



from functools import wraps

import pymc as pm

from .GP import *


class GPC(GP):
    @wraps(GP.build_model)
    def build_model(
        self,
        seed=None,
        continuous_kernel="ExpQuad",
        heteroskedastic_inputs=False,
        heteroskedastic_outputs=False,
        sparse=False,
        n_u=100,
        eps=1e-6,
    ):
        if heteroskedastic_inputs:
            raise NotImplementedError("The GP Classifier does not support heteroskedastic inputs.")
        if heteroskedastic_outputs:
            raise NotImplementedError("The GP Classifier does not support heteroskedastic outputs.")
        if sparse:
            raise NotImplementedError("The GP Classifier does not support sparse structure (yet).")

        self.build_latent(seed=seed, continuous_kernel=continuous_kernel, eps=eps)

        _, y = self.get_shaped_data("mean")

        with self.model:
            f = self.prior

            # logit link and Bernoulli likelihood
            p = pm.Deterministic("p", pm.math.invlogit(f))
            _ = pm.Bernoulli("y", p=p, observed=y)

        return self

    @wraps(GP.draw_point_samples)
    def draw_point_samples(
        self,
        points,
        *args,
        source=None,
        output=None,
        var_name="posterior_samples",
        additive_level="total",
        increment_var=True,
        **kwargs
    ):
        var_name = self._recursively_append(var_name, increment_var=increment_var)

        # A
        self.stdzr.logit_vars += [var_name]

        return super(GPC, self).draw_point_samples(
            points,
            *args,
            source=source,
            output=output,
            var_name=var_name,
            additive_level=additive_level,
            increment_var=True,
            **kwargs
        )



import pathlib as pl

import pandas as pd
import pymc as pm
import pytest
import seaborn as sns

from gumbi import GP, DataSet, ParrayPlotter, Standardizer

test_dir = pl.Path(__file__).resolve().parent
test_data = test_dir / "test_data"

example_stdzr = {
    "a": {"μ": -0.762, "σ2": 1.258**2},
    "b": {"μ": -0.0368, "σ2": 0.351**2},
    "c": {"μ": -5.30, "σ2": 0.582**2},
    "d": {"μ": -0.307, "σ2": 0.158**2},
    "e": {"μ": -1.056, "σ2": 0.398**2},
    "f": {"μ": 3.34, "σ2": 0.1501**2},
    "X": {"μ": -0.282, "σ2": 1**2},
    "Y": {"μ": 4.48, "σ2": 0.75**2},
    "lg10_Z": {"μ": 5, "σ2": 2**2},
}

log_vars = ["d", "f", "b", "c", "Y"]
logit_vars = ["e", "X"]

################################################################################
# GP
################################################################################


@pytest.fixture
def example_estimates():
    es = pd.read_pickle(test_data / "test_dataset.pkl")
    stdzr = Standardizer(**example_stdzr, log_vars=log_vars, logit_vars=logit_vars)
    ds = DataSet.from_tidy(es, names_column="Parameter", stdzr=stdzr)
    return ds


@pytest.fixture
def example_gp(example_estimates):
    return GP(example_estimates, outputs="d")


# Input argument parsing


def test_gp_default_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=["X", "Y"])
    assert gp.continuous_dims == ["X", "Y"]
    assert gp.categorical_dims == []


def test_gp_data_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=["X", "Y"])
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 2
    assert y.shape == (66,)


def test_gp_numerical_continuous_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=["X", "Y", "lg10_Z"])
    assert gp.continuous_dims == ["X", "Y", "lg10_Z"]
    for dim in gp.continuous_dims:
        assert len(gp.continuous_levels[dim]) == len(gp.data.tidy[dim].unique())
        assert len(gp.continuous_coords[dim].values()) == len(gp.continuous_levels[dim])
    assert gp.categorical_dims == []
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 3
    assert y.shape == (66,)


def test_gp_categorical_continuous_fit_parsing(example_gp):
    gp = example_gp.specify_model(continuous_dims=["X", "Y", "Name"])
    assert gp.continuous_dims == ["X", "Y", "Name"]
    for dim in gp.continuous_dims:
        assert len(gp.continuous_levels[dim]) == len(gp.data.tidy[dim].unique())
        assert len(gp.continuous_coords[dim].values()) == len(gp.continuous_levels[dim])
    assert gp.categorical_dims == []
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 3
    assert y.shape == (66,)


def test_gp_params_fit_parsing(example_gp):
    gp = example_gp.specify_model(outputs=["d", "c"], continuous_dims=["X", "Y"])
    assert gp.continuous_dims == ["X", "Y"]
    assert gp.categorical_dims == ["Parameter"]
    assert gp.categorical_levels == {"Parameter": ["d", "c"]}
    assert gp.categorical_coords == {"Parameter": {"d": 1, "c": 0}}
    X, y = gp.get_structured_data()
    assert X.shape == (66,)
    assert len(X.names) == 2
    assert y.shape == (66,)
    assert len(y.names) == 2


def test_gp_single_input_fit_parsing(example_gp):
    gp = example_gp.specify_model(
        continuous_dims=["X", "Y", "Name"],
        continuous_levels={"Name": ["intense-opportunity"]},
    )
    assert gp.continuous_dims == ["X", "Y"]
    assert gp.filter_dims == {"Name": ["intense-opportunity"], "Parameter": ["d"]}
    X, y = gp.get_structured_data()
    assert X.shape == (7,)
    assert len(X.names) == 2
    assert y.shape == (7,)


# Model building

# def test_gp_build_model_simple(example_gp):
#     gp = example_gp.specify_model(continuous_dims=['X', 'Y'])
#     gp.build_model()
#     assert isinstance(gp.model, pm.model.Model)
#     assert isinstance(gp.gp_dict['total'], pm.gp.gp.Marginal)


# Combinatorial gp objects with various parameterizations
@pytest.fixture(params=[False, True])
def additive(request):
    return request.param


@pytest.fixture(
    params=[
        {"outputs": ["d", "c"], "continuous_dims": ["X", "Y"]},
        {"continuous_dims": ["X", "Y"], "categorical_dims": "Code"},
        {"continuous_dims": ["X", "Y", "Name"]},
        {"continuous_dims": ["X", "Y", "lg10_Z"]},
        {
            "continuous_dims": ["X", "Y", "Name"],
            "continuous_levels": {"Name": ["intense-opportunity"]},
        },
    ]
)
def fit_inputs(request):
    return request.param


@pytest.mark.slow
def test_gp_build_model(example_gp, fit_inputs, additive):
    # Basically just makes sure that `build_model` runs without errors
    gp = example_gp.specify_model(**fit_inputs, additive=additive)
    gp.build_model()
    assert isinstance(gp.model, pm.model.Model)
    assert isinstance(gp.gp_dict["total"], pm.gp.gp.Marginal)


@pytest.mark.slow
def test_gp_build_model_additive(example_gp):
    # Basically just makes sure that `build_model` runs without errors
    gp = example_gp.specify_model(
        outputs=["d", "c"],
        continuous_dims=["X", "Y"],
        categorical_dims="lg10_Z",
        additive=True,
    )
    gp.build_model()
    assert isinstance(gp.model, pm.model.Model)
    assert all(name in gp.gp_dict.keys() for name in ["total", "global", "lg10_Z"])


# MAP estimation


def test_gp_fit_simple(example_gp):
    # Basically just makes sure that `fit` runs without errors
    gp = example_gp.fit(continuous_dims=["X", "Y", "lg10_Z"], continuous_levels={"lg10_Z": [8]})
    assert isinstance(gp.MAP, dict)


@pytest.mark.slow
def test_gp_fit(example_gp, fit_inputs, additive):
    # Basically just makes sure that `fit` runs without errors
    gp = example_gp.fit(**fit_inputs, additive=additive)
    assert isinstance(gp.MAP, dict)

def test_gp_predict():
    cars = sns.load_dataset('mpg').dropna()
    ds = DataSet(cars, outputs=['mpg', 'acceleration'], log_vars=['mpg', 'acceleration', 'horsepower'])
    gp = GP(ds)
    gp.fit(outputs=['mpg'], continuous_dims=['horsepower']);
    X = gp.prepare_grid()
    y = gp.predict_grid()
    ParrayPlotter(X, y).plot()
    
    


"""Gaussian Process Model Building Interface"""

# from gumbi import data, style, utils

from . import data, style, utils
from .aggregation import *
from .arrays import *
from .array_utils import *
from .plotting import *
from .regression import *
from .versions import __version__

# Aliases
parray = ParameterArray
uarray = UncertainArray
uparray = UncertainParameterArray
mvuparray = MVUncertainParameterArray



import warnings

import numpy as np
import pymc as pm
import pytensor.tensor as pt

from gumbi.aggregation import DataSet
from gumbi.arrays import (  # noqa: F401
    MVUncertainParameterArray,
    ParameterArray,
    UncertainParameterArray,
)
from gumbi.utils.gp_utils import get_ls_prior
from gumbi.utils.misc import assert_in

from ..base import Regressor

__all__ = ["GP"]


class GP(Regressor):
    r"""Gaussian Process surface learning and prediction.

    See Also
    --------
    :class:`Regressor`

    Notes
    -----
    The GP class is built from a dataframe in the form of a :class:`DataSet`
    object. The output(s) are taken from :attr:`DataSet.outputs` and the
    corresponding column in the tidy data frame taken from
    :attr:`DataSet.names_column`. This column will be generically referred to as
    the `output_column` in this documentation, but can take any value specifed
    when the :class:`DataSet` is constructed. The model inputs are constructed
    by filtering this dataframe, extracting column values, and converting these
    to numerical input coordinates. The main entry point will be :meth:`fit`,
    which parses the dimensions of the model with :meth:`specify_model`,
    extracts numerical input coordinates with :meth:`get_structured_data`,
    compiles the Pymc model with :meth:`build_model`, and finally learns the
    hyperparameters with :meth:`find_MAP`.

    Dimensions fall into several categories:

    * Filter dimensions, those with only one level, are used to subset the
      dataframe but are not included as explicit inputs to the model. These are
      not specified explicitly, but rather any continuous or categorical
      dimension with only one level is treated as a filter dimension.

    * Continuous dimensions are treated as explicit coordinates and given a
      Radial Basis Function kernel

      * Linear dimensions (which must be a subset of `continuous_dims`) have an
        additional linear kernel.

    * Coregion dimensions imply a distinct but correlated output for each level

      * If more than one output is specified, the `output_column` is treated as
        a categorical dim.

    A non-additive model has the form:

    .. math::

        y &\sim \text{Normal} \left( \mu, \sigma \right) \\ mu &\sim
        \mathcal{GP} \left( K \right) \\ K &= \left( K^\text{cont}+K^\text{lin}
        \right) K^\text{coreg}_\text{outputs} \prod_{n} K^\text{coreg}_{n} \\
        K^\text{cont} &= \text{RBF} \left( \ell_{i}, \eta \right) \\
        K^\text{lin} &= \text{LIN} \left( c_{j}, \tau \right) \\ K^\text{coreg}
        &= \text{Coreg} \left( \boldsymbol{W}, \kappa \right) \\ \sigma &\sim
        \text{Exponential} \left( 1 \right) \\

    Where :math:`i` denotes a continuous dimension, :math:`j` denotes a linear
    dimension, and :math:`n` denotes a categorical dimension (excluding the
    `output_column`). :math:`K^\text{cont}` and :math:`K^\text{lin}` each
    consist of a joint kernel encompassing all continuous and linear dimensions,
    respectively, whereas :math:`K^\text{coreg}_{n}` is a distinct kernel for a
    given categorical dimension.

    The additive model has the form:

    .. math::

        mu &\sim \mathcal{GP}\left( K^\text{global} \right) + \sum_{n}
        \mathcal{GP}\left( K_{n} \right) \\ K^\text{global} &= \left(
        K^\text{cont}+K^\text{lin} \right) K^\text{coreg}_\text{outputs} \\
        K_{n} &= \left( K^\text{cont}_{n}+K^\text{lin}_{n} \right)
        K^\text{coreg}_\text{outputs} K^\text{coreg}_{n} \\

    Note that, in the additive model, :math:`K^\text{cont}_{n}` and
    :math:`K^\text{lin}_{n}` still consist of only the continuous and linear
    dimensions, respectively, but have unique priors corresponding to each
    categorical dimension. However, there is only one
    :math:`K^\text{coreg}_\text{outputs}` kernel.

    Parameters
    ----------
    dataset : DataSet
        Data for fitting.
    outputs : str or list of str, default None
        Name(s) of output(s) to learn. If ``None``, uses all values from
        ``outputs`` attribute of *dataset*.
    seed : int
        Random seed

    Examples
    --------
    A GP object is created from a :class:`DataSet` and can be fit immediately
    with the default dimension configuration (regressing `r` with RBF + linear
    kernels for `X` and `Y`):

    >>> import gumbi as gmb
    >>> df = pd.read_pickle(gmb.data.example_dataset)
    >>> outputs=['a', 'b', 'c', 'd', 'e', 'f']
    >>> ds = gmb.DataSet(df, outputs=outputs, log_vars=['Y', 'b', 'c', 'd', 'f'], logit_vars=['X', 'e'])
    >>> gp = gmb.GP(ds, outputs='d').fit()

    Note that last line is equivalent to

    >>> gp = gmb.GP(ds, outputs='d')
    >>> gp.specify_model()
    >>> gp.build_model()
    >>> gp.find_MAP()

    The model can be specified with various continuous, linear, and categorical
    dimensions. `X` and `Y` are always included in both ``continuous_dims`` and
    ``linear_dims``.

    >>> gp.specify_model(continuous_dims='lg10_Z', linear_dims='lg10_Z', categorical_dims='Pair')
    >>> gmb.GP(ds).fit(continuous_dims='lg10_Z', linear_dims='lg10_Z', categorical_dims='Pair')  # equivalent

    After the model is fit, define a grid of points at which to make
    predictions. The result is a :class:`ParameterArray`:

    >>> gp.prepare_grid()
    >>> gp.grid_points
    ('X', 'Y'): [(0.075     ,  10.) (0.08358586,  10.) (0.09217172,  10.) ...
     (0.90782828, 800.) (0.91641414, 800.) (0.925     , 800.)]

    Make predictions, returning an :class:`UncertainParameterArray`

    >>> gp.predict_grid()
    >>> gp.predictions
    d['μ', 'σ2']: [[(0.70728056, 0.16073197) (0.70728172, 0.16073197)
                    (0.70728502, 0.16073197) ... (0.70727954, 0.16073197)
                    (0.7072809 , 0.16073197) (0.70728058, 0.16073197)]
                   ...
                   [(0.70749247, 0.1607318 ) (0.70773573, 0.16073116)
                    (0.70806603, 0.16072949) ... (0.70728449, 0.16073197)
                    (0.70728194, 0.16073197) (0.7072807 , 0.16073197)]]

    The `uparray` makes it easy to calculate standard statistics in natural or
    transformed/standardized space while maintaining the original shape of the
    array:

    >>> gp.predictions.z.dist.ppf(0.025)
    array([[-3.1887916 , -3.18878491, -3.18876601, ..., -3.18879744,
            -3.18878966, -3.18879146],
           ...,
           [-3.1875742 , -3.18617286, -3.18426272, ..., -3.18876906,
            -3.18878366, -3.18879081]])

    Finally, plot the results:

    >>> import matplotlib.pyplot as plt
    >>>
    >>> plt.style.use(str(gmb.style.futura))
    >>> x_pa = gp.predictions_X
    >>> y_upa = gp.predictions
    >>> gmb.ParrayPlotter(x_pa, y_upa).plot()

    Plot a slice down the center of the prediction along each axis

    >>> x_pa, y_upa = gp.get_conditional_prediction(Y=88)
    >>>
    >>> ax = gmb.ParrayPlotter(x_pa, y_upa).plot()
    >>> ax.set_xticklabels([int(float(txt.get_text())*100) for txt in ax.get_xticklabels()]);

    Plot a slice down the center of the prediction along each axis

    >>> x_pa, y_upa = gp.get_conditional_prediction(X=0.5)
    >>>
    >>> ax = gmb.ParrayPlotter(x_pa, y_upa).plot()
    >>> ax.set_xticklabels([int(float(txt.get_text())*100) for txt in ax.get_xticklabels()]);


    Attributes
    ----------
    dataset : DataSet
        Data for fitting.
    outputs : list of str
        Name(s) of output(s) to learn.
    seed : int
        Random seed
    continuous_dims : list of str
        Columns of dataframe used as continuous dimensions
    linear_dims : list of str
        Subset of continuous dimensions to apply an additional linear kernel.
    continuous_levels : dict
        Values considered within each continuous column as ``{dim: [level1,
        level2]}``
    continuous_coords : dict
        Numerical coordinates of each continuous level within each continuous
        dimension as ``{dim: {level: coord}}``
    categorical_dims : list of str
        Columns of dataframe used as categorical dimensions
    categorical_levels : dict
        Values considered within each categorical column as ``{dim: [level1,
        level2]}``
    categorical_coords : dict
        Numerical coordinates of each categorical level within each categorical
        dimension as ``{dim: {level: coord}}``
    additive : bool
        Whether to treat categorical dimensions as additive or joint
    filter_dims : dict
        Dictionary of column-value pairs used to filter dataset before fitting
    X : array
        A 2D tall array of input coordinates.
    y : array
        A 1D vector of observations
    model : pymc.model.Model
        Compiled pymc model
    gp_dict : dict
        Dictionary of model GP objects. Contains at least 'total'.
    """

    def __init__(self, dataset: DataSet, outputs=None, seed=2021):
        super(GP, self).__init__(dataset, outputs, seed)

        self.model = None
        self.gp_dict = None
        self.MAP = None
        self.trace = None

        self.continuous_kernel = "ExpQuad"
        self.heteroskedastic_inputs = False
        self.heteroskedastic_outputs = True
        self.sparse = False
        self.latent = False
        self.n_u = 100

        self.model_specs = {
            "seed": self.seed,
            "continuous_kernel": self.continuous_kernel,
            "heteroskedastic_inputs": self.heteroskedastic_inputs,
            "heteroskedastic_outputs": self.heteroskedastic_outputs,
            "sparse": self.sparse,
            "n_u": self.n_u,
        }

    ################################################################################
    # Model building and fitting
    ################################################################################

    def fit(
        self,
        outputs=None,
        linear_dims=None,
        continuous_dims=None,
        continuous_levels=None,
        continuous_coords=None,
        categorical_dims=None,
        categorical_levels=None,
        additive=False,
        seed=None,
        continuous_kernel="ExpQuad",
        period=None,
        heteroskedastic_inputs=False,
        heteroskedastic_outputs=True,
        sparse=False,
        n_u=100,
        ARD=True,
        ls_bounds=None,
        mass=0.98,
        spec_kwargs=None,
        build_kwargs=None,
        MAP_kwargs=None,
    ):
        """Fits a GP surface

        Parses inputs, compiles a Pymc model, then finds the MAP value for the
        hyperparameters. `{}_dims` arguments indicate the columns of the
        dataframe to be included in the model, with `{}_levels` indicating which
        values of those columns are to be included (``None`` implies all
        values).

        If ``additive==True``, the model is constructed as the sum of a global
        GP and a distinct GP for each categorical dimension. Each of these GPs,
        including the global GP, consists of an RBF+linear kernel multiplied by
        a coregion kernel for the `output_column` if necessary. Although the
        same continuous kernel structure is used for each GP in this model,
        unique priors are assigned to each distinct kernel. However, there is
        always only one coregion kernel for the `output_column`. The kernel for
        each dimension-specific GP is further multiplied by a coregion kernel
        that provides an output for each level in that dimension.

        See Also
        --------
        :meth:`build_model`

        Parameters
        ----------
        outputs : str or list of str, default None
            Name(s) of output(s) to learn. If ``None``, :attr:`outputs` is used.
        linear_dims : str or list of str, optional
            Subset of continuous dimensions to apply an additional linear
            kernel. If ``None``, defaults to ``['Y','X']``.
        continuous_dims : str or list of str, optional
            Columns of dataframe used as continuous dimensions.
        continuous_levels : str, list, or dict, optional
            Values considered within each continuous column as ``{dim: [level1,
            level2]}``.
        continuous_coords : list or dict, optional
            Numerical coordinates of each continuous level within each
            continuous dimension as ``{dim: {level: coord}}``.
        categorical_dims : str or list of str, optional
            Columns of dataframe used as categorical dimensions.
        categorical_levels : str, list, or dict, optional
            Values considered within each categorical column as ``{dim: [level1,
            level2]}``.
        additive : bool, default False
            Whether to treat categorical_dims as additive or joint (default).
        seed : int, optional.
            Random seed for model instantiation. If ``None``, :attr:`seed` is
            used.
        continuous_kernel : {'ExpQuad', 'Matern32', 'Matern52', 'Exponential',
        or 'Cosine'}
            Covariance function to use for continuous dimensions. See `pymc
            docs`_ for more details.
        period : ParameterArray, optional
            A single parray of length 1 with one layer for each
            `continuous_dims` by name containing the period of the kernel, if
            periodic-like kernel is used.
        heteroskedastic_inputs: bool, default False
            Whether to allow heteroskedasticity along continuous dimensions
            (input-dependent noise)
        heteroskedastic_outputs: bool, default True
            Whether to allow heteroskedasticity between multiple outputs
            (output-dependent noise). `Not yet implemented`
        sparse: bool, default False
            Whether to use a `sparse approximation`_ to the GP.
        n_u: int, default 100
            Number of inducing points to use for the sparse approximation, if
            required.
        ARD: bool, default True
            Whether to use "Automatic Relevance Determination" in the continuous
            kernel. If _True_, each continuous dimension receives its own
            lengthscale; otherwise a single lengthscale is used for all
            continuous dimensions.

        **MAP_kwargs
            Additional keyword arguments passed to :func:`pm.find_MAP`.

        Returns
        -------
        self : :class:`GP`
        """

        self.specify_model(
            outputs=outputs,
            linear_dims=linear_dims,
            continuous_dims=continuous_dims,
            continuous_levels=continuous_levels,
            continuous_coords=continuous_coords,
            categorical_dims=categorical_dims,
            categorical_levels=categorical_levels,
            additive=additive,
            **(spec_kwargs or {}),
        )

        self.build_model(
            seed=seed,
            continuous_kernel=continuous_kernel,
            period=period,
            heteroskedastic_inputs=heteroskedastic_inputs,
            heteroskedastic_outputs=heteroskedastic_outputs,
            sparse=sparse,
            n_u=n_u,
            ARD=ARD,
            ls_bounds=ls_bounds,
            mass=mass,
            **(build_kwargs or {}),
        )

        self.find_MAP(**(MAP_kwargs or {}))

        return self

    def _make_continuous_cov(
        self,
        continuous_cov_func,
        D_in,
        idx_s,
        n_s,
        ls_params,
        ARD=True,
        period=None,
        **kernel_kwargs,
    ):
        shape = n_s if ARD else 1

        if period is not None:
            zperiods = [period.z[dim + "_z"].values() for dim in self.continuous_dims]
            kernel_kwargs["period"] = np.array(zperiods).squeeze()

        def continuous_cov(suffix):
            ls = pm.InverseGamma(f"ls_{suffix}", **ls_params, shape=shape)
            # ls = pm.Gamma(f"ls_{suffix}", alpha=2, beta=1, shape=shape)
            η = pm.Gamma(f"η_{suffix}", alpha=2, beta=1)
            cov = η**2 * continuous_cov_func(input_dim=D_in, active_dims=idx_s, ls=ls, **kernel_kwargs)

            return cov

        return continuous_cov

    def _make_periodic_cov(
        self,
        continuous_cov_func,
        D_in,
        idx_s,
        n_s,
        ls_params,
        period,
        ARD=True,
        **kernel_kwargs,
    ):
        continuous_kernel_factory = self._make_continuous_cov(
            continuous_cov_func, 2, None, n_s, ls_params, ARD=ARD, **kernel_kwargs
        )

        zperiods = [period.z[dim + "_z"].values() for dim in self.continuous_dims]
        zperiods = np.array(zperiods).squeeze() if len(zperiods) > 1 else zperiods[0]

        def mapping(x, zperiods):
            c = 2.0 * np.pi * (1.0 / zperiods)
            u = pt.concatenate((pt.sin(c * x), pt.cos(c * x)), 1)
            return u

        def periodic_cov(suffix):
            cov_func = continuous_kernel_factory(suffix)
            cov = pm.gp.cov.WarpedInput(
                input_dim=D_in, active_dims=idx_s, cov_func=cov_func, warp_func=mapping, args=(zperiods,)
            )

            return cov

        return periodic_cov

    def _make_linear_cov(self, D_in, idx_l, n_l):
        def linear_cov(suffix):
            c = pm.Normal(f"c_{suffix}", mu=0, sigma=10, shape=n_l)
            τ = pm.HalfNormal(f"τ_{suffix}", sigma=10)
            return τ * pm.gp.cov.Linear(input_dim=D_in, c=c, active_dims=idx_l)

        return linear_cov

    def _make_coreg_cov(self, D_in, seed):
        def coreg_cov(suffix, D_out, idx):
            initval = np.random.default_rng(seed).standard_normal(size=(D_out, 2))
            W = pm.Normal(f"W_{suffix}", mu=0, sigma=3, shape=(D_out, 2), initval=initval)
            κ = pm.Gamma(f"κ_{suffix}", alpha=1.5, beta=1, shape=(D_out,))
            return pm.gp.cov.Coregion(input_dim=D_in, active_dims=[idx], kappa=κ, W=W)

        return coreg_cov

    # TODO: add full probabilistic model description to docstring TODO: allow dimension-specific continuous kernel
    # specification TODO: allow single multi-dimensional continuous kernel rather than independent kernels per dimension
    def build_model(
        self,
        seed=None,
        continuous_kernel="ExpQuad",
        period=None,
        heteroskedastic_inputs=False,
        heteroskedastic_outputs=True,
        sparse=False,
        n_u=100,
        ARD=True,
        ls_bounds=None,
        mass=0.98,
    ):
        r"""Compile a marginalized pymc model for the GP.

        Each dimension in :attr:`continuous_dims` is combined in an ExpQuad kernel with a principled
        :math:`\text{InverseGamma}` prior for each lengthscale (as `suggested by Michael Betancourt`_) and a
        :math:`\text{Gamma}\left(2, 1\right)` prior for variance.

        .. _suggested by Michael Betancourt: https://betanalpha.github.io/assets/case_studies/gp_part3/part3.html#4_adding_an_informative_prior_for_the_length_scale  # noqa: E501
        .. _pymc docs: https://docs.pymc.io/en/v3/api/gp/cov.html
        .. _sparse approximation: https://docs.pymc.io/en/v3/pymc-examples/examples/gaussian_processes/GP-SparseApprox.html  # noqa: E501

        Parameters
        ----------
        seed : int, optional.
            Random seed. If ``None``, :attr:`seed` is used.
        continuous_kernel : {'ExpQuad', 'Matern32', 'Matern52', 'Exponential', 'Cosine', or 'Periodic'}
            Covariance function to use for continuous dimensions. See `pymc docs`_ for more details.
        period : ParameterArray, optional
            A single parray of length 1 with one layer for each `continuous_dims` by name containing the period of the kernel, if periodic-like kernel is used.
        heteroskedastic_inputs: bool, default False
            Whether to allow heteroskedasticity along continuous dimensions (input-dependent noise).
        heteroskedastic_outputs: bool, default True
            Whether to allow heteroskedasticity between multiple outputs (output-dependent noise). `Not yet
            implemented`.
        sparse: bool, default False
            Whether to use a `sparse approximation`_ to the GP.
        n_u: int, default 100
            Number of inducing points to use for the sparse approximation, if required.
        ARD: bool, default True
            Whether to use "Automatic Relevance Determination" in the continuous kernel. If _True_, each continuous
            dimension receives its own lengthscale; otherwise a single lengthscale is used for all continuous
            dimensions.

        Returns
        -------
        self : :class:`GP`
        """

        if heteroskedastic_inputs:
            raise NotImplementedError("Heteroskedasticity over inputs is not yet implemented.")

        X, y = self.get_shaped_data("mean")
        D_in = len(self.dims)
        assert X.shape[1] == D_in

        seed = self.seed if seed is None else seed
        self.seed = seed
        self.continuous_kernel = continuous_kernel
        self.heteroskedastic_inputs = heteroskedastic_inputs
        self.heteroskedastic_outputs = heteroskedastic_outputs
        self.sparse = sparse
        self.n_u = n_u
        self.latent = False

        self.model_specs = {
            "seed": seed,
            "continuous_kernel": continuous_kernel,
            "heteroskedastic_inputs": heteroskedastic_inputs,
            "heteroskedastic_outputs": heteroskedastic_outputs,
            "sparse": sparse,
            "n_u": n_u,
        }
        gp_dict = self._construct_kernels(
            X,
            continuous_kernel,
            seed,
            sparse,
            latent=False,
            ARD=ARD,
            ls_bounds=ls_bounds,
            mass=mass,
            period=period,
        )

        with self.model:
            # From https://docs.pymc.io/notebooks/GP-Marginal.html OR a covariance function for the noise can be given
            # noise_l = pm.Gamma("noise_l", alpha=2, beta=2) cov_func_noise = pm.gp.cov.Exponential(1, noise_l) +
            # pm.gp.cov.WhiteNoise(sigma=0.1) y_ = gp.marginal_likelihood("y", X=X, y=y, sigma=cov_func_noise)

            # GP is heteroskedastic across outputs by default, but homoskedastic across continuous dimensions
            σ = pm.Exponential("σ", lam=1)
            noise = pm.gp.cov.WhiteNoise(sigma=σ)
            if heteroskedastic_inputs:
                raise NotImplementedError("Heteroskedasticity over inputs is not yet implemented")
                # noise += continuous_cov('noise')
            if heteroskedastic_outputs and self.out_col in self.categorical_dims:
                D_out = len(self.categorical_levels[self.out_col])
                coreg_cov = self._make_coreg_cov(D_in, seed)
                idx_p = self._get_dim_indexes()["p"]
                noise *= coreg_cov("Output_noise", D_out, idx_p)

            if sparse:
                Xu = pm.gp.util.kmeans_inducing_points(n_u, X)
                if heteroskedastic_outputs:
                    warnings.warn(
                        "Heteroskedasticity over outputs is not yet implemented for sparse GP. \
                            Reverting to scalar-valued noise."
                    )
                _ = gp_dict["total"].marginal_likelihood("ml", X=X, Xu=Xu, y=y, sigma=σ)
            else:
                _ = gp_dict["total"].marginal_likelihood("ml", X=X, y=y, sigma=noise)

        # self.gp_dict = gp_dict
        return self

    def _choose_implementation(self, sparse=False, latent=False):
        if sparse and latent:
            raise NotImplementedError("Sparse Latent GPs are not yet implemented.")

        if sparse:
            pm_gp = pm.gp.MarginalSparse
            gp_kws = {"approx": "FITC"}
        elif latent:
            pm_gp = pm.gp.Latent
            gp_kws = {}
        else:
            pm_gp = pm.gp.Marginal
            gp_kws = {}

        def implementation(*args, **kwargs):
            return pm_gp(*args, **{**kwargs, **gp_kws})  # Fix once Python >= 3.9

        return implementation

    def _get_dim_counts(self):
        dim_counts = {
            "l": len(self.linear_dims),
            "s": len(self.continuous_dims),
            "c": len(self.categorical_dims),
            "p": len(self.outputs),
        }

        return dim_counts

    def _get_dim_indexes(self):
        dim_indexes = {
            "l": [self.dims.index(dim) for dim in self.linear_dims],
            "s": [self.dims.index(dim) for dim in self.continuous_dims],
            "c": [self.dims.index(dim) for dim in self.categorical_dims],
            "p": self.dims.index(self.out_col) if self.out_col in self.dims else None,
        }

        return dim_indexes

    # def _prepare_lengthscales(self, X, ARD=False):
    #     X_s = X[:, self._get_dim_indexes()["s"]]

    #     ls_μ, ls_σ = [stat for stat in np.array([get_ls_prior(dim) for dim in X_s.T]).T]
    #     return ls_μ, ls_σ

    def _prepare_lengthscales(self, X, *, ARD, ls_bounds=None, mass=0.98):
        X_s = X[:, self._get_dim_indexes()["s"]]
        
        if ls_bounds is not None:
            zbounds = [
                [b if not np.isnan(b) else None for b in ls_bounds[dim].z.values().squeeze()]
                # if dim in ls_bounds.names else (None, None)
                for dim in self.continuous_dims
                if dim in ls_bounds.names
            ]
            lower, upper = list(zip(*zbounds))
            if not ARD and len(lower) != 1 or len(upper) != 1:
                raise ValueError("Bounds must be specified for only a single dimension if ARD is False")
        else:
            lower, upper = None, None
            
        ls_params = get_ls_prior(X_s, ARD=ARD, lower=lower, upper=upper, mass=mass)

        # ls_μ, ls_σ = [stat for stat in np.array([get_ls_prior(dim) for dim in X_s.T]).T]
        # return ls_μ, ls_σ
        return ls_params

    def _construct_kernels(
        self,
        X,
        continuous_kernel,
        seed,
        sparse,
        latent,
        ARD=True,
        ls_bounds=None,
        mass=0.98,
        period=None,
    ):
        continuous_kernels = [
            "ExpQuad",
            # "RatQuad",
            "Matern12",
            "Matern32",
            "Matern52",
            "Exponential",
            "Periodic",
        ]
        continuous_kernels += [kernel + "+Periodic" for kernel in continuous_kernels if kernel != "Periodic"]
        assert_in("Continuous kernel", continuous_kernel, continuous_kernels)

        if continuous_kernel.endswith("+Periodic"):
            continuous_kernel = continuous_kernel.removesuffix("+Periodic")
            if period is None:
                raise ValueError("Period must be specified for periodic kernel")
            kernel_factory = self._make_periodic_cov
        else:
            kernel_factory = self._make_continuous_cov

        continuous_cov_func = getattr(pm.gp.cov, continuous_kernel)

        D_in = len(self.dims)

        ns = self._get_dim_counts()
        idxs = self._get_dim_indexes()
        ls_params = self._prepare_lengthscales(X, ARD=ARD, ls_bounds=ls_bounds, mass=mass)

        continuous_cov = kernel_factory(
            continuous_cov_func,
            D_in,
            idxs["s"],
            ns["s"],
            ls_params,
            ARD=ARD,
            period=period,
        )
        linear_cov = self._make_linear_cov(D_in, idxs["l"], ns["l"])
        coreg_cov = self._make_coreg_cov(D_in, seed)

        pm_gp = self._choose_implementation(sparse=sparse, latent=latent)

        with pm.Model() as self.model:
            # μ = pm.Normal('μ', mu=0, sigma=10) β = pm.Normal('β', mu=0, sigma=10, shape=n_l) lin_mean =
            # pm.gp.mean.Linear(coeffs=[β[i] if i in idx_l else 0 for i in range(D_in), intercept=μ)

            # Define a "global" continuous kernel regardless of additive structure
            cov = continuous_cov("total")
            if ns["l"] > 0:
                cov += linear_cov("total")

            # Construct a coregion kernel for each categorical_dims
            if ns["c"] > 0 and not self.additive:
                for dim, idx in zip(self.categorical_dims, idxs["c"]):
                    if dim == self.out_col:
                        continue
                    D_out = len(self.categorical_levels[dim])
                    cov *= coreg_cov(dim, D_out, idx)

            # Coregion kernel for parameters, if necessary
            if self.out_col in self.categorical_dims:
                D_out = len(self.categorical_levels[self.out_col])
                cov_param = coreg_cov(self.out_col, D_out, idxs["p"])
                cov *= cov_param

            gp_dict = {"total": pm_gp(cov_func=cov)}

            # Construct a continuous+coregion kernel for each categorical_dim, then combine them additively
            if self.additive:
                gp_dict["global"] = gp_dict["total"]
                for dim, idx in zip(self.categorical_dims, idxs["c"]):
                    if dim == self.out_col:
                        continue

                    # Continuous kernel specific to this dimension
                    cov = continuous_cov(dim)
                    # TODO: Decide if each additive dimension needs its own linear kernel
                    if ns["l"] > 0:
                        cov += linear_cov(dim)

                    # Coregion kernel specific to this dimension
                    D_out = len(self.categorical_levels[dim])
                    cov *= coreg_cov(dim, D_out, idx)

                    # Coregion kernel for parameters, if necessary
                    if self.out_col in self.categorical_dims:
                        cov *= cov_param

                    # Combine GPs
                    gp_dict[dim] = pm_gp(cov_func=cov)
                    gp_dict["total"] += gp_dict[dim]

        self.gp_dict = gp_dict
        return gp_dict

    def build_latent(
        self,
        seed=None,
        continuous_kernel="ExpQuad",
        prior_name="latent_prior",
        ARD=True,
        lower=None,
        upper=None,
        mass=0.98,
    ):
        if self.additive:
            raise NotImplementedError("Additive/latent GPs are not yet implemented")

        X, y = self.get_shaped_data("mean")
        D_in = len(self.dims)
        assert X.shape[1] == D_in

        seed = self.seed if seed is None else seed
        self.seed = seed
        self.continuous_kernel = continuous_kernel
        self.sparse = False
        self.latent = True

        gp_dict = self._construct_kernels(
            X,
            continuous_kernel,
            seed,
            sparse=False,
            latent=True,
            ARD=ARD,
            lower=lower,
            upper=upper,
            mass=mass,
        )

        with self.model:
            self.prior = gp_dict["total"].prior(prior_name, X=X)

        return self

    def find_MAP(self, *args, **kwargs):
        """Finds maximum a posteriori value for hyperparameters in model.

        Parameters
        ----------
        *args
            Positional arguments passed to :func:`pm.find_MAP`
        **kwargs
            Keyword arguments passed to :func:`pm.find_MAP`
        """
        assert self.model is not None
        with self.model:
            self.MAP = pm.find_MAP(*args, **kwargs)

        return self.MAP

    def sample(self, *args, **kwargs):
        """Draws samples from the posterior for the hyperparameters in model.

        Parameters
        ----------
        *args
            Positional arguments passed to :func:`pm.sample`
        **kwargs
            Keyword arguments passed to :func:`pm.sample`
        """

        defaults = {
            "return_inferencedata": True,
            "random_seed": self.seed,
        }

        assert self.model is not None
        with self.model:
            self.trace = pm.sample(*args, **{**defaults, **kwargs})

        return self.trace

    def predict(self, points_array, with_noise=True, additive_level="total", **kwargs):

        # TODO: need to supply "given" dict for additive GP sublevel predictions
        if additive_level != "total":
            raise NotImplementedError("Prediction for additive sublevels is not yet supported.")

        with self.model:
            # Prediction means and variance as a numpy vector
            predictions = self.gp_dict[additive_level].predict(
                points_array, point=self.MAP, diag=True, pred_noise=with_noise, **kwargs
            )

        return predictions

    def _recursively_append(self, var_name, suffix="_", increment_var=True):
        if var_name in self.model.named_vars:
            if increment_var:
                var_name += suffix
                return self._recursively_append(var_name)
            else:
                raise ValueError(f'The variable name "{var_name}" already exists in model.')
        else:
            return var_name

    def draw_point_samples(
        self,
        points,
        *args,
        source=None,
        output=None,
        var_name="posterior_samples",
        additive_level="total",
        increment_var=True,
        **kwargs,
    ):
        """Draw posterior samples at supplied points

        Parameters
        ----------
        points : ParameterArray
            1-D ParameterArray vector of coordinates for prediction, must have one layer per ``self.dims``
        output : str or list of str, optional
            Variable for which to make predictions
        source : {None, dict, az.data.inference_data.InferenceData}
            GP parameters for which to draw samples. Should be the result of :meth:`find_MAP`, :meth:`sample`, or
            _None_.
        var_name : str, default "posterior_samples"
            Name to assign new variable to contain conditional predictions.
        additive_level : str, default "total"
            Level of additive GP at which to make predictions.
        increment_var : bool, default True
            Whether to append '_' to the end of _var_name_ if it already exists in model.

        Returns
        -------
        samples : parray
            Samples as a 'Parray'
        """

        output = self._parse_prediction_output(output)
        if len(output) > 1:
            raise NotImplementedError("Drawing correlated samples of multiple outputs is not yet implemented.")
        points_array, tall_points, param_coords = self._prepare_points_for_prediction(points, output=output)

        if source is None:
            if self.trace is None and self.MAP is None:
                raise ValueError('"Source" of predictions must be supplied if GP object has no trace or MAP stored.')
            elif self.trace is not None and self.MAP is not None:
                raise ValueError('"Source" of predictions must be supplied if GP object has both trace and MAP stored.')
            elif self.MAP is not None:
                source = [self.MAP]
            elif self.trace is not None:
                source = self.trace

        var_name = self._recursively_append(var_name, increment_var=increment_var)

        with self.model:
            _ = self.gp_dict[additive_level].conditional(var_name, points_array)

        with self.model:
            samples = pm.sample_posterior_predictive(*args, source, var_names=[var_name], **kwargs)

        self.predictions = self.parray(**{output: samples[var_name]}, stdzd=True)
        self.predictions_X = points

        return self.predictions

    def draw_grid_samples(
        self,
        *args,
        source=None,
        output=None,
        categorical_levels=None,
        var_name="posterior_samples",
        additive_level="total",
        increment_var=True,
        **kwargs,
    ):
        """Draw posterior samples at points defined by :meth:`prepare_grid`.

        Parameters
        ----------
        source : {None, dict, az.data.inference_data.InferenceData}
            GP parameters for which to draw samples. Should be the result of :meth:`find_MAP`, :meth:`sample`, or
            _None_.
        output : str or list of str, optional
            Variable(s) for which to make predictions
        categorical_levels : dict, optional
            Level for each :attr:`categorical_dims` at which to make prediction
        var_name : str, default "posterior_samples"
            Name to assign new variable to contain conditional predictions.
        additive_level : str, default "total"
            Level of additive GP at which to make predictions.
        increment_var : bool, default True
            Whether to append '_' to the end of _var_name_ if it already exists in model.

        Returns
        -------
        samples : ParameterArray
            Samples as a 'Parray' reshaped into a grid with _len(:attr:`continuous_dims`)_ dimensions
        """

        if self.grid_points is None:
            raise ValueError("Grid must first be specified with `prepare_grid`")

        points = self.grid_points
        if self.categorical_dims:
            points = self.append_categorical_points(points, categorical_levels=categorical_levels)

        samples = self.draw_point_samples(
            *args,
            points=points,
            output=output,
            source=source,
            var_name=var_name,
            additive_level=additive_level,
            increment_var=increment_var,
            **kwargs,
        )
        self.predictions = samples.reshape(-1, *self.grid_parray.shape)
        self.predictions_X = self.predictions_X.reshape(self.grid_parray.shape)

        return self.predictions



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import pathlib as pl
import sys

sys.path.insert(0, os.path.abspath("../../gumbi/"))

PROJECT_ROOT = pl.Path(__file__).resolve().parent.parent.parent
VERSION = PROJECT_ROOT / "VERSION"

# -- Project information -----------------------------------------------------

project = "Gumbi"
copyright = "2021, John Goertz"
author = "John Goertz"

# The full version, including alpha/beta/rc tags
with open(VERSION, encoding="utf-8") as f:
    version = f.read().strip()
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    # 'sphinxcontrib.rsvgconverter',
    # 'numpydoc',
    "sphinx.ext.napoleon",
    # 'sphinx_copybutton',
    "sphinx_rtd_theme",
    "nbsphinx",
]
# numpydoc_show_class_members = True
numpydoc_class_members_toctree = False

autosummary_generate = True

# autodoc_default_options = {
#     'members': True,
#     'special-members': True,
#     'private-members': False,
#     'inherited-members': True,
#     'undoc-members': False,
#     'exclude-members': '__weakref__',
# }


nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc=figure.dpi=96",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_css_files = [
    "green_theme.css",
]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



from scipy.spatial.distance import pdist
import numpy as np
import pymc as pm
from .misc import listify, first
from warnings import warn


def parse_ls_limits(X, *, ARD, lower=None, upper=None):
    if ARD:
        all_points = np.hsplit(X, X.shape[1])
    else:
        all_points = [X]

    lowers = [None] if lower is None else listify(lower)
    if len(lowers) == 1:
        lowers = lowers * len(all_points)
    if len(lowers) != len(all_points):
        raise ValueError("Number of lower bounds must match number of dimensions")

    uppers = [None] if upper is None else listify(upper)
    if len(uppers) == 1:
        uppers = uppers * len(all_points)
    if len(uppers) != len(all_points):
        raise ValueError("Number of upper bounds must match number of dimensions")

    for i, (points, lower, upper) in enumerate(zip(all_points, lowers, uppers)):
        distances = pdist(points)
        distinct = distances != 0
        
        default_lower = distances[distinct].min() if sum(distinct) > 0 else 0.01

        if lower is None:
            lower = default_lower
        lower = max(lower, default_lower, 0.01)
        if upper is None:
            upper = distances[distinct].max() if sum(distinct) > 0 else 1

        lowers[i] = lower
        uppers[i] = upper

    return lowers, uppers


def get_ls_prior(X, *, ARD, lower=None, upper=None, mass=0.98):
    
    lowers, uppers = parse_ls_limits(X, ARD=ARD, lower=lower, upper=upper)

    params = []

    for i, (lower, upper) in enumerate(zip(lowers, uppers)):

        converged = False
        mass_ = mass
        while not converged:
            try:
                params_ls = pm.find_constrained_prior(
                    distribution=pm.InverseGamma,
                    lower=lower,
                    upper=upper,
                    init_guess={"alpha": lower, "beta": upper},
                    mass=mass_,
                )
            except ValueError as e:
                if 'Optimization of parameters failed' in str(e):
                    mass_ -= 0.01
                else:
                    raise e
            else:
                converged = True
                if mass_ != mass:
                    warn("Mass of constrained lengthscale prior was "
                         f"reduced from {mass:.3f} to {mass_:.3f} to enable "
                         f"convergence for dimension {i}.")

        params.append(params_ls)

    params = {k: [dim[k] for dim in params] for k in first(params).keys()}

    return params


# def get_ls_prior(points):
#     distances = pdist(points[:, None])
#     distinct = distances != 0
#     ls_l = distances[distinct].min() if sum(distinct) > 0 else 0.1
#     ls_u = distances[distinct].max() if sum(distinct) > 0 else 1
#     ls_σ = max(0.1, (ls_u - ls_l) / 6)
#     ls_μ = ls_l + 3 * ls_σ
#     return ls_μ, ls_σ



"""Internal utility functions, not intended for public consumption"""

from functools import wraps
from itertools import chain, islice
from typing import Iterable, Iterator
from operator import attrgetter

import numpy as np


class NotExactlyOneError(Exception):
    """Raised when more than one value is found in a list."""

    pass


def one(itr: Iterable):
    """Ensures list has exactly one element, then returns it."""
    if isinstance(itr, (set, Iterator)):
        itr = list(itr)
    if (n := len(itr)) != 1:
        raise NotExactlyOneError(f"Expected one element in list, got {n}")
    return listify(itr)[0]


def first(itr: Iterable):
    """Returns first element in list."""
    return listify(itr)[0]


def extract(attr, itr):
    """Extracts attribute from each element in iterable."""
    return list(map(attrgetter(attr), itr))


def listify(x):
    """Wraps input in a list if it's not already one, or converts an iterable."""
    if isinstance(x, list):
        return x
    elif isinstance(x, str):
        return [x]
    elif isinstance(x, (set, Iterator, Iterable)):
        return list(x)
    elif x is None:
        return []
    else:
        return [x]


def flatten(list_of_lists, depth=-1):
    """Flatten specified number of levels of nesting. If depth is -1, flatten all."""
    match depth:
        case 0:
            return list_of_lists
        case -1:
            if not isinstance(first(list_of_lists), list):
                return list_of_lists
            else:
                depth = 0
        case _:
            pass
    flatter = list(chain.from_iterable(list_of_lists))
    return flatten(flatter, depth - 1)


def group_by(itr, key, unique=False):
    """Groups elements in iterable by key."""
    if isinstance(key, str):
        key = attrgetter(key)
    if unique:
        one_ = one
    else:
        one_ = skip
    groups = {group: one_([el for el in itr if key(el) == group]) for group in set(map(key, itr))}
    return groups


def skip(x):
    """Returns input unmodified."""
    return x


def NotImplementedWrapper(func):
    """Blocks execution and raises an error."""

    @wraps(func)
    def block(*args, **kwargs):
        """Raises error"""
        raise NotImplementedError

    return block


def assert_in(name: str, arg, itr: Iterable):
    """Raises error if value not in list."""
    if arg not in itr:
        raise ValueError(f"{name} must be one of {itr}")


def assert_is_subset(name: str, subset: Iterable, superset: Iterable):
    """Raises error if any required value not in list."""
    l_set = set(superset)
    r_set = set(subset)
    if not l_set.issuperset(r_set):
        missing = list(r_set.difference(l_set))
        msg = f"{list_is_are(missing)} missing from {name}"
        raise ValueError(msg)


def assert_one(names: str, itr: Iterable):
    """Raises error if none or multiple of values in `itr` are not None."""
    if sum(el is not None for el in itr) != 1:
        raise ValueError(f"Exactly one of {names} must be supplied")


def list_is_are(lst: list) -> str:
    """String formats list to be grammatically correct."""
    lst = listify(lst)
    if len(lst) == 0:
        return None
    elif len(lst) == 1:
        msg = f"{list_and(lst)} is"
    else:
        msg = f"{list_and(lst)} are"
    return msg


def list_and(lst: list) -> str:
    """String formats list to insert 'and' before last element."""
    lst = listify(lst)
    if len(lst) == 0:
        msg = ""
    elif len(lst) == 1:
        msg = f"{lst[0]}"
    elif len(lst) == 2:
        msg = f"{lst[0]} and {lst[1]}"
    else:
        msg = f'{", ".join(lst[:-1])}, and {lst[-1]}'
    return msg


def s(n):
    return "s" if n != 1 else ""


def round_to_n(x, n=2):
    """Round to a specified number of significant digits."""
    if isinstance(x, float):
        rounded = np.round(x, -int(np.floor(np.log10(np.abs(x))) - (n - 1)))
    elif isinstance(x, (list, np.ndarray)):
        rounded = [(np.round(x_, -int(np.floor(np.log10(np.abs(x_))) - (n - 1))) if x_ != 0 else 0) for x_ in x]
    else:
        raise ValueError("x must be float, list, or ndarray.")
    return np.where(x == 0.0, x, rounded)


def prettyprint_dict(dct, lpad=2):
    # Determine left-padding by length of longest key
    m = max(map(len, list(dct.keys()))) + lpad
    lines = []
    for k, v in dct.items():
        left = k.rjust(m)
        if isinstance(v, str):
            right = v
        else:
            right = np.array2string(np.array(v), prefix=(k.rjust(m) + ": "))
        line = f"{left}: {right}"
        lines.append(line)
    return "\n".join(lines)


def batched(iterable, n):
    """Yield successive n-sized batches from iterable."""
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


class Trigger(object):
    """Descriptor to create a property that calls an instance method when its value is changed.
    Note, must be specified as a class variable.

    Parameters
    ----------
    method: str
        The name of the instance method to be called
    default: optional
        Default value of variable
    """

    def __init__(self, method, default=None):
        self.default = default
        self.method = method
        self.data = dict()

    def __get__(self, instance, owner):
        return self.data.get(instance, self.default)

    def __set__(self, instance, value):
        self.data[instance] = value
        getattr(instance, self.method)()


class InstanceCopy:
    """A class that creates instances by copying all attributes of a parent instance"""

    def __init__(self, parent):
        assert isinstance(parent, self.__class__.__bases__[-1])

        for attr in parent.__dict__:
            setattr(self, attr, getattr(parent, attr))



"""
Instructions for building the package.

$ python -m build
$ twine check dist/*
$ twine upload dist/gumbi-$(cat VERSION)*

"""

import pathlib as pl

from setuptools import find_packages, setup

DISTNAME = "gumbi"
DESCRIPTION = "Gaussian Process Model Building Interface"
AUTHOR = "John Goertz"
AUTHOR_EMAIL = ""
URL = "https://github.com/JohnGoertz/Gumbi"
LICENSE = "Apache 2.0"

PROJECT_ROOT = pl.Path(__file__).resolve().parent
REQUIREMENTS = PROJECT_ROOT / "requirements.txt"
README = PROJECT_ROOT / "README.md"
VERSION = PROJECT_ROOT / "VERSION"

with open(REQUIREMENTS) as f:
    install_reqs = f.read().splitlines()

with open(README, "r") as fh:
    long_description = fh.read()

with open(VERSION, encoding="utf-8") as f:
    version = f.read().strip()

classifiers = [
    "Development Status :: 4 - Beta",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Mathematics",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: OS Independent",
]

setup(
    name=DISTNAME,
    version=version,
    author="John Goertz",
    author_email="",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    url=URL,
    license=LICENSE,
    python_requires=">=3.9",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_reqs,
    classifiers=classifiers,
    # keywords=['python'],
)



"""Parses current package version"""

import pkg_resources

try:
    __version__ = str(pkg_resources.get_distribution("gumbi").parsed_version)
except pkg_resources.DistributionNotFound:
    __version__ = "develop"
